/************************************************************************
 *   IRC - Internet Relay Chat, include/common.h
 *   Copyright (C) 1990 Armin Gruner
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef	__common_include__
#define __common_include__

#include <time.h>

#ifdef	PARAMH
#include <sys/param.h>
#endif

#ifndef PROTO
#if __STDC__
#	define PROTO(x)	x
#else
#	define PROTO(x)	()
#endif
#endif

#ifndef NULL
#define NULL 0
#endif

#ifdef TRUE
#undef TRUE
#endif

#ifdef FALSE
#undef FALSE
#endif

#define FALSE (0)
#define TRUE  (!FALSE)

#ifndef UNSURE
#define UNSURE (2)
#endif

#if 0
#ifndef	MALLOCH
char	*malloc(), *calloc();
void	free();
#else
#include MALLOCH
#endif
#endif

extern	char 	*collapse PROTO((char *));
extern	int	match PROTO((char *, char *));
#define mycmp(a,b) \
 ( (toupper((a)[0])!=toupper((b)[0])) || smycmp((a)+1,(b)+1) )
extern  int     smycmp PROTO((char *, char *));
extern	int	myncmp PROTO((char *, char *, int));
#ifdef NEED_STRTOK
extern	char	*strtok PROTO((char *, char *));
#endif
#ifdef NEED_STRTOKEN
extern	char	*strtoken PROTO((char **, char *, char *));
#endif
#ifdef NEED_INET_ADDR
extern unsigned long inet_addr PROTO((char *));
#endif

#if defined(NEED_INET_NTOA) || defined(NEED_INET_NETOF) && !defined(SCOUNIX)
#include <netinet/in.h>
#endif

int global_count, max_global_count;
extern char *myctime PROTO((time_t));
extern char *strtoken PROTO((char **, char *, char *));

#ifndef MAX
#define MAX(a, b)	((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a, b)	((a) < (b) ? (a) : (b))
#endif

#define DupString(x,y) do{x=MyMalloc(strlen(y)+1);(void)strcpy(x,y);}while(0)

#ifndef SCOUNIX
extern u_char tolowertab[], touppertab[];
#else
extern unsigned char tolowertab[], touppertab[]; 
	/* Typedef for u_char exists and is wrong. */
#endif

#undef tolower
#define tolower(c) (tolowertab[(u_char)(c)])

#undef toupper
#define toupper(c) (touppertab[(u_char)(c)])

#undef isalpha
#undef isdigit
#undef isxdigit
#undef isalnum
#undef isprint
#undef isascii
#undef isgraph
#undef ispunct
#undef islower
#undef isupper
#undef isspace
#undef iscntrl

extern unsigned char char_atribs[];

#define PRINT 1
#define CNTRL 2
#define ALPHA 4
#define PUNCT 8
#define DIGIT 16
#define SPACE 32
#define ALLOW 64

#ifndef KLINE_TEMP
#define KLINE_PERM 0
#define KLINE_TEMP 1
#define KLINE_AKILL 2
#endif

#define	iscntrl(c) (char_atribs[(u_char)(c)]&CNTRL)
#define isallowed(c) (char_atribs[(u_char)(c)]&ALLOW)
#define isalpha(c) (char_atribs[(u_char)(c)]&ALPHA)
#define isspace(c) (char_atribs[(u_char)(c)]&SPACE)
#define islower(c) ((char_atribs[(u_char)(c)]&ALPHA) && ((u_char)(c) > 0x5f))
#define isupper(c) ((char_atribs[(u_char)(c)]&ALPHA) && ((u_char)(c) < 0x60))
#define isdigit(c) (char_atribs[(u_char)(c)]&DIGIT)
#define	isxdigit(c) (isdigit(c) || 'a' <= (c) && (c) <= 'f' || \
		     'A' <= (c) && (c) <= 'F')
#define isalnum(c) (char_atribs[(u_char)(c)]&(DIGIT|ALPHA))
#define isprint(c) (char_atribs[(u_char)(c)]&PRINT)
#define isascii(c) ((u_char)(c) >= 0 && (u_char)(c) <= 0x7f)
#define isgraph(c) ((char_atribs[(u_char)(c)]&PRINT) && ((u_char)(c) != 0x32))
#define ispunct(c) (!(char_atribs[(u_char)(c)]&(CNTRL|ALPHA|DIGIT)))

extern char *MyMalloc();
extern void flush_connections();
extern struct SLink *find_user_link();

/*
 * Protocol support text.  DO NO CHANGE THIS unless you know what
 * you are doing.
 */
#define COMPAT_LEVEL 1
#ifdef IPV6
#define PROTOCTL_SUPPORTED "NOQUIT TOKEN WATCH=128 SAFELIST IPV6"
#else
#define PROTOCTL_SUPPORTED "NOQUIT TOKEN WATCH=128 SAFELIST"
#endif

#if defined(__FreeBSD__) || defined(__APPLE__)
extern char *malloc_options;
#endif

extern int lu_noninv, lu_inv, lu_serv, lu_oper, 
           lu_unknown, lu_channel, lu_lu, lu_lulocal, lu_lserv, 
           lu_clu, lu_mlu, lu_cglobalu, lu_mglobalu;
           
extern time_t	now;

/*
 IPv6 Support by Jeroen Massar <jeroen@unfix.org> aka Fuzzel / unfix
 Most differences in the code use an #ifdef IPV6 to differentiate between IPv4/IPv6
*/

#ifdef IPV6

#include <netinet/in.h>


#if defined(linux) || defined(__NetBSD__)
#ifndef s6_laddr
#define s6_laddr			s6_addr32
#endif
#endif
 
#if defined(__OpenBSD__) || defined(__FreeBSD__)
#define s6_laddr			__u6_addr.__u6_addr32
#include <sys/socket.h>
#endif

#define AFINET				AF_INET6
#define SOCKADDR_IN		sockaddr_in6
#define SOCKADDR			sockaddr_in6
#define SIN_FAMILY		sin6_family
#define SIN_PORT			sin6_port
#define SIN_ADDR			sin6_addr
#define S_ADDR				s6_addr
#define S_NET				s6_net
#define S_HOST				s6_host
#define S_LH				s6_lh
#define IN_ADDR			in6_addr
#define SOCK_LEN_TYPE	size_t
 
#define MYDUMMY_SIZE 128
char mydummy[MYDUMMY_SIZE];
char mydummy2[MYDUMMY_SIZE];

#define AND16(x) ((x)[0]&(x)[1]&(x)[2]&(x)[3]&(x)[4]&(x)[5]&(x)[6]&(x)[7]&(x)[8]&(x)[9]&(x)[10]&(x)[11]&(x)[12]&(x)[13]&(x)[14]&(x)[15])
static unsigned char minus_one[]={ 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 0};
#define WHOSTENTP(x) ((x)[0]|(x)[1]|(x)[2]|(x)[3]|(x)[4]|(x)[5]|(x)[6]|(x)[7]|(x)[8]|(x)[9]|(x)[10]|(x)[11]|(x)[12]|(x)[13]|(x)[14]|(x)[15])

#if defined(linux)
static const struct in6_addr in6addr_any={ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0};
#endif

#else /* Support "old" IPv4 too :) */

#define AFINET				AF_INET
#define SOCKADDR_IN		sockaddr_in
#define SOCKADDR			sockaddr
#define SIN_FAMILY		sin_family
#define SIN_PORT			sin_port
#define SIN_ADDR			sin_addr
#define S_ADDR				s_addr
#define S_NET				s_net
#define S_HOST				s_host
#define S_LH				s_lh
#define IN_ADDR			in_addr
#define SOCK_LEN_TYPE	int

#define WHOSTENTP(x) (x)
 
#endif /* IPV6 */

#ifndef SCOUNIX /* SCO Openserver has these, but the header files are broken */
#ifdef NEED_INET_NTOA
extern char *inet_ntoa PROTO((struct IN_ADDR));
#endif

#ifdef NEED_INET_NETOF
extern int inet_netof PROTO((struct IN_ADDR));
#endif
#endif /* SCOUNIX */

#endif /* __common_include__ */

/*
 * $Log: common.h,v $
 * Revision 1.5  2001/01/16 03:48:35  smkelly
 * Removed all WIN32 code
 *
 * Revision 1.4  2000/12/13 10:22:24  smkelly
 * Removed some _WIN32 code
 *
 * Revision 1.3  2000/10/25 18:12:57  smkelly
 * Removed CASETABLEs
 *
 * Revision 1.2  1999/11/12 11:12:07  smkelly
 * Added RCS tags
 *
 */

