/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_auth.c
 *   Copyright (C) 1992 Darren Reed
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "patchlevel.h"
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#if defined(__hpux)
#include "inet.h"
#endif
#include <fcntl.h>
#include "sock.h"		/* If FD_ZERO isn't define up to this point,  */
			/* define it (BSD4.2 needs this) */
#include "h.h"


CVSID("$Id: s_set.c,v 1.21 2001/01/16 03:48:38 smkelly Exp $");

#define TOKEN_BAD		-1
#define TOKEN_SHOWALL		0
#define TOKEN_RESET		1
#define TOKEN_LOCKED		2
#define TOKEN_CONNTIMEOUT 	3
#define TOKEN_AUTOCONN		4
#define TOKEN_MAXCHANS		5
#define TOKEN_KILLCHASETIME	6
#define TOKEN_NICKTHROTTLE	7
#define TOKEN_SOCKSCHECK	8
#define TOKEN_CONNINFO		9
#define TOKEN_ADMINSET		10
#define TOKEN_MASTERKEY		11
#define TOKEN_NOSPOOF		12
#define TOKEN_AUTOINVIS		13
#define TOKEN_AUTOCLOAK		14
#define TOKEN_FLOOD		15
#define TOKEN_MOTDFILE		16
#define TOKEN_INFOFILE		17
#define TOKEN_NONEWCHANS	18

static char *set_table[] = {
	"SHOWALL",
	"RESET",
	"LOCKED",
	"CONNTIMEOUT",
	"AUTOCONN",
	"MAXCHANS",
	"KILLCHASETIME",
	"NICKTHROTTLE",
	"SOCKSCHECK",
	"CONNINFO",
	"ADMINSET",
	"MASTERKEY",
	"NOSPOOF",
	"AUTOINVIS",
	"AUTOCLOAK",
	"FLOOD",
	"MOTDFILE",
	"INFOFILE",
	"NONEWCHANS",
	NULL
};

int set_parser(char *pt)
{
	int i;
	
	for (i=0; set_table[i]; i++)
		if (!strcasecmp(pt, set_table[i]))
			return i;
	return TOKEN_BAD;
}

int m_set(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int parc;
char *parv[];
{
	int cnum;
	char *cmd;

	if (!MyClient(sptr) || (S_ADMINSET && !IsAdmin(sptr))) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	if (parc > 1) {
		cmd = parv[1];
		cnum = set_parser(cmd);
		switch (cnum) {
		case TOKEN_RESET:
			sendto_realops("%s has issued a SET RESET", parv[0]);
			initialize_global_set_options(1);
			return 0;
			break;
		case TOKEN_SHOWALL:
			parc = 1;
		case TOKEN_LOCKED:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
				if (nv == 0) {
				   sendto_realops("%s has disabled LOCKED mode",
				   	parv[0]);
				   S_LOCKED = 0;
				} else {
				   sendto_realops("%s has enabled LOCKED mode",
				   	parv[0]);
				   S_LOCKED = 1;
				}
			} else {
				sendto_one(sptr,
				":%s NOTICE %s :LOCKED is currently %s",
				me.name, parv[0], S_LOCKED?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_CONNTIMEOUT:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv < 30 || nv > 90) {
			      sendto_one(sptr,
			      ":%s NOTICE %s :CONNTIMEOUT must be between 30 and 90",
			      me.name, parv[0]);
			      return 0;
			   }
			   sendto_realops("%s has set CONNTIMEOUT to %d seconds",
			   	parv[0], nv);
			   CONNECTTIMEOUT = nv;
			   return 0;
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :CONNTIMEOUT is currently set at %d seconds",
			   me.name, parv[0], CONNECTTIMEOUT);
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_AUTOCONN:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv == 0) {
				   sendto_realops("%s has disabled AUTOCONN mode",
				   	parv[0]);
				   S_AUTOCONN = 0;
			   } else {
				   sendto_realops("%s has enabled AUTOCONN mode",
				   	parv[0]);
				   S_AUTOCONN = 1;
			   }
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :AUTOCONN is currently %s",
			   me.name, parv[0], S_AUTOCONN?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_MAXCHANS:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv < 0 || nv > 50) {
			   	sendto_one(sptr,
			   	":%s NOTICE %s :MAXCHANS must be between 0 and 50",
			   	me.name, parv[0]);
			   	return 0;
			   }
			   sendto_realops("%s has set MAXCHANS to %d", parv[0], nv);
			   MAXCHANNELSPERUSER = nv;
			   return 0;
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :MAXCHANS is currently set at %d",
			   me.name, parv[0], MAXCHANNELSPERUSER);
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_KILLCHASETIME:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv < 0 || nv > 3600) {
			      sendto_one(sptr,
			      ":%s NOTICE %s :KILLCHASETIME must be between 0 and 3600",
			      me.name, parv[0]);
			      return 0;
			   }
			   sendto_realops("%s has set KILLCHASETIME to %d seconds",
			   	parv[0], nv);
			   KILLCHASETIMELIMIT = nv;
			   return 0;
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :KILLCHASETIME is currently set at %d seconds",
			   me.name, parv[0], KILLCHASETIMELIMIT);
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_NICKTHROTTLE:
			if (parc > 3) {
			   int nc = atoi(parv[2]);
			   int nd = atoi(parv[3]);
			   if (nc < 0 || nc > 100) {
			   	sendto_one(sptr,
			   	":%s NOTICE %s :NICKTHROTTLE: First argument must be between 0 and 100",
			   	me.name, parv[0]);
			   	return 0;
			   } else if (nd < 0 || nd > 3600) {
			   	sendto_one(sptr,
			   	":%s NOTICE %s :NICKTHROTTLE: Second argument must be between 0 and 3600",
			   	me.name, parv[0]);
			   	return 0;
			   }
			   sendto_realops("%s has set NICKTHROTTLE to %d changes per %d seconds",
			   	parv[0], nc, nd);
			   NICKCOUNT = nc;
			   NICKDELAY = nd;
			   return 0;
			} else if (parc > 2) {
			   sendto_one(sptr,
			   ":%s NOTICE %s :NICKTHROTTLE: Not enough arguments.  Need count and delay",
			   me.name, parv[0]);
			   return 0;
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :NICKTHROTTLE is currently set at %d changes per %d seconds",
			   me.name, parv[0], NICKCOUNT, NICKDELAY);
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_SOCKSCHECK:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv == 0) {
				   sendto_realops("%s has disabled SOCKSCHECK",
				   	parv[0]);
				   S_SOCKSCHECK = 0;
			   } else {
				   sendto_realops("%s has enabled SOCKSCHECK",
				   	parv[0]);
				   S_SOCKSCHECK = 1;
			   }
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :SOCKSCHECK is currently %s",
			   me.name, parv[0], S_SOCKSCHECK?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_CONNINFO:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv == 0) {
				   sendto_realops("%s has disabled CONNINFO",
				   	parv[0]);
				   SHOWCONNECTINFO = 0;
			   } else {
				   sendto_realops("%s has enabled CONNINFO",
				   	parv[0]);
				   SHOWCONNECTINFO = 1;
			   }
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :CONNINFO is currently %s",
			   me.name, parv[0], SHOWCONNECTINFO?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_ADMINSET:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (!IsAdmin(sptr)) {
			   	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
			   	return 0;
			   }
			   
			   if (nv == 0) {
				   sendto_realops("%s has disabled ADMINSET",
				   	parv[0]);
				   S_ADMINSET = 0;
			   } else {
				   sendto_realops("%s has enabled ADMINSET",
				   	parv[0]);
				   S_ADMINSET = 1;
			   }
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :ADMINSET is currently %s",
			   me.name, parv[0], S_ADMINSET?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_MASTERKEY:
			if (parc > 2) {
			   u_char *s1, *s2;
			   parv[2] = check_string(parv[2]);
			   for (s1 = s2 = (u_char *)parv[2]; *s2; s2++)
			   	if ((*s1 = *s2 & 0x7f) > (u_char) 32 &&	*s1 != ':')
			   		s1++;
			   	*s1 = '\0';
			   if (BadPtr(parv[2])) return 0;
			   strncpy(S_MASTERKEY, parv[2], KEYLEN);
			   sendto_realops("%s set MASTERKEY to %s",
			   	parv[0], parv[2]);
			} else {
			   if (!BadPtr(S_MASTERKEY))
				sendto_one(sptr,
				":%s NOTICE %s :MASTERKEY is currently set to \"%s\"",
				me.name, parv[0], S_MASTERKEY);
			   else
			   	sendto_one(sptr,
			   	":%s NOTICE %s :MASTERKEY is currently not set",
			   	me.name, parv[0]);
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_NOSPOOF:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv == 0) {
				   sendto_realops("%s has disabled NOSPOOF",
				   	parv[0]);
				   NOSPOOF = 0;
			   } else {
				   sendto_realops("%s has enabled NOSPOOF",
				   	parv[0]);
				   NOSPOOF = 1;
			   }
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :NOSPOOF is currently %s",
			   me.name, parv[0], NOSPOOF?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;		
		case TOKEN_AUTOINVIS:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv == 0) {
				   sendto_realops("%s has disabled AUTOINVIS",
				   	parv[0]);
				   S_AUTOINVIS = 0;
			   } else {
				   sendto_realops("%s has enabled AUTOINVIS",
				   	parv[0]);
				   S_AUTOINVIS = 1;
			   }
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :AUTOINVIS is currently %s",
			   me.name, parv[0], S_AUTOINVIS?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;	
		case TOKEN_AUTOCLOAK:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv == 0) {
				   sendto_realops("%s has disabled AUTOCLOAK",
				   	parv[0]);
				   S_AUTOCLOAK = 0;
			   } else {
				   sendto_realops("%s has enabled AUTOCLOAK",
				   	parv[0]);
				   S_AUTOCLOAK = 1;
			   }
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :AUTOCLOAK is currently %s",
			   me.name, parv[0], S_AUTOCLOAK?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_FLOOD:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv < 512 || nv > 8000) {
			      sendto_one(sptr,
			      ":%s NOTICE %s :FLOOD must be between 512 and 8000",
			      me.name, parv[0]);
			      return 0;
			   }
			   sendto_realops("%s has set FLOOD to %d bytes",
			   	parv[0], nv);
			   CLIENT_FLOOD = nv;
			   return 0;
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :FLOOD is currently set at %d bytes",
			   me.name, parv[0], CLIENT_FLOOD);
			}
			if (cnum != TOKEN_SHOWALL) return 0;		
		case TOKEN_MOTDFILE:
			if (parc > 2) {
			   u_char *chk;
			   parv[2] = check_string(parv[2]);
   			   for (chk = (u_char *)parv[2]; *chk; chk++)
			   	if (*chk == '/') {
			   		sendto_one(sptr,
			   			":%s NOTICE %s :MOTDFILE must not contain a path",
			   			me.name, parv[0]);
			   		return 0;
			   	}
			   if (BadPtr(parv[2])) return 0;
			   strncpy(MOTD, parv[2], 1024);
			   sendto_realops("%s set MOTDFILE to %s",
			   	parv[0], parv[2]);
			} else {
				sendto_one(sptr,
				":%s NOTICE %s :MOTDFILE is currently set to \"%s\"",
				me.name, parv[0], MOTD);
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_INFOFILE:
			if (parc > 2) {
			   u_char *chk;
			   parv[2] = check_string(parv[2]);
   			   for (chk = (u_char *)parv[2]; *chk; chk++)
			   	if (*chk == '/') {
			   		sendto_one(sptr,
			   			":%s NOTICE %s :INFOFILE must not contain a path",
			   			me.name, parv[0]);
			   		return 0;
			   	}
			   if (BadPtr(parv[2])) return 0;
			   strncpy(INFO, parv[2], 1024);
			   sendto_realops("%s set INFOFILE to %s",
			   	parv[0], parv[2]);
			} else {
				sendto_one(sptr,
				":%s NOTICE %s :INFOFILE is currently set to \"%s\"",
				me.name, parv[0], INFO);
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		case TOKEN_NONEWCHANS:
			if (parc > 2) {
			   int nv = atoi(parv[2]);
			   if (nv == 0) {
				   sendto_realops("%s has disabled NONEWCHANS mode",
				   	parv[0]);
				   S_NONEWCHANS = 0;
			   } else {
				   sendto_realops("%s has enabled NONEWCHANS mode",
				   	parv[0]);
				   S_NONEWCHANS = 1;
			   }
			} else {
			   sendto_one(sptr,
			   ":%s NOTICE %s :NONEWCHANS is currently %s",
			   me.name, parv[0], S_NONEWCHANS?"enabled":"disabled");
			}
			if (cnum != TOKEN_SHOWALL) return 0;
		default:
		case TOKEN_BAD:
			if (cnum == TOKEN_SHOWALL) return 0;
			break;
		}
	}
	for (cnum=0; set_table[cnum]; cnum++)
		sendto_one(sptr, ":%s NOTICE %s :Options: %s",
			me.name, parv[0], set_table[cnum]);
	return 0;
}


/*
 * $Log: s_set.c,v $
 * Revision 1.21  2001/01/16 03:48:38  smkelly
 * Removed all WIN32 code
 *
 * Revision 1.20  2000/12/02 12:40:37  smkelly
 * SET NONEWCHANS added
 *
 * Revision 1.19  2000/10/25 18:14:51  smkelly
 * Fixed void return in non-void m_set()
 *
 * Revision 1.18  2000/10/25 07:19:56  smkelly
 * Added SET MOTDFILE/INFOFILE
 *
 * Revision 1.17  2000/07/23 21:40:25  smkelly
 * SET FLOOD
 *
 * Revision 1.16  2000/07/19 20:17:37  smkelly
 * SET AUTOINVIS, SET AUTOCLOAK
 *
 * Revision 1.15  2000/07/19 10:28:01  smkelly
 * Cosmetic tweak to SET MASTERKEY
 *
 * Revision 1.14  2000/07/19 10:26:11  smkelly
 * SET NOSPOOF
 *
 * Revision 1.13  2000/07/19 08:21:34  smkelly
 * SET MASTERKEY
 *
 * Revision 1.12  2000/07/19 06:38:15  smkelly
 * SET RESET
 *
 * Revision 1.11  2000/07/19 05:16:11  smkelly
 * SET ADMINSET
 *
 * Revision 1.10  2000/07/19 04:52:04  smkelly
 * SET now only sends to real IRC Operators
 *
 * Revision 1.9  2000/07/19 04:35:46  smkelly
 * SET SHOWALL and SET CONNINFO
 *
 * Revision 1.8  2000/07/19 04:03:32  smkelly
 * SET SOCKSCHECK
 *
 * Revision 1.7  2000/07/19 03:35:53  smkelly
 * UNIXPORT support removed
 *
 * Revision 1.6  2000/07/19 00:23:48  smkelly
 * SET NICKTHROTTLE
 *
 * Revision 1.5  2000/07/18 22:32:21  smkelly
 * SET KILLCHASETIME
 *
 * Revision 1.4  2000/07/18 21:49:37  smkelly
 * SET MAXCHANS
 *
 * Revision 1.3  2000/07/18 20:35:25  smkelly
 * SET AUTOCONN
 *
 * Revision 1.2  2000/07/18 19:46:47  smkelly
 * SET CONNTIMEOUT
 *
 * Revision 1.1  2000/07/18 10:10:27  smkelly
 * New SET command with LOCKED mode
 *
 *
 */

